from .web3_module import (  # noqa: F401
    Web3ModuleTest,
)
from .eth_module import (  # noqa: F401
    EthModuleTest,
)
from .net_module import (  # noqa: F401
    NetModuleTest,
)
from .personal_module import (  # noqa: F401
    PersonalModuleTest,
)
from .version_module import (  # noqa: F401
    VersionModuleTest,
)

from .parity_module import (  # noqa: F401
    ParityModuleTest,
)
